#include "unit.h"

#include <stdlib.h> // abs()

unit::unit(struct model * model, float x, float y) :
    energy(new gauge(model->max_energy)),
    cristal(new gauge(model->max_cristal)),
    powergem(new gauge(model->max_powergem))
{
    this->model = model;
    this->x = x;
    this->y = y;
    this->target_x = 0.0;
    this->target_y = 0.0;
    this->energy_accumulator = 0;
}

unit::~unit() {
    delete(this->cristal);
    delete(this->energy);
}

void unit::draw(SDL_Renderer * renderer) {
    extern struct sprite * icon_cristal;
    extern struct sprite * icon_energy;
    extern struct sprite * icon_powergem;
    this->model->sprite.draw(renderer, this->x, this->y);
    this->powergem->draw_top(renderer, this, icon_powergem, 0);
    this->cristal->draw_top(renderer, this, icon_cristal, this->powergem->value);
    this->energy->draw_bot(renderer, this, icon_energy, 0);
}
bool unit::can_build(struct model * model) {
    return(this->energy->value  >= model->energy_cost
        && this->cristal->value >= model->cristal_cost);
}

void unit::set_position(float x, float y) {
    if(x >= 0 && x < SCREEN_WIDTH && y >= 0 && y < SCREEN_HEIGHT) {
        this->x = x;
        this->y = y;
    }
}

void unit::move(float dx, float dy, struct square grid[GRID_WIDTH][GRID_HEIGHT]) {
    if(this->model->speed > 0) {
        float nx = this->x + dx * this->model->speed;
        float ny = this->y + dy * this->model->speed;
        int mx = nx/SQUARE_SIZE;
        int my = ny/SQUARE_SIZE;
        if(grid[mx][my].is_passable()) {
            this->set_position(nx, ny);
        } else if(grid[mx][(int)this->y/SQUARE_SIZE].is_passable()) {
            this->set_position(nx, this->y);
        } else if(grid[(int)this->x/SQUARE_SIZE][my].is_passable()) {
            this->set_position(this->x, ny);
        }
    }
}

void unit::move_toward(float x, float y, struct square grid[GRID_WIDTH][GRID_HEIGHT]) {
    float X = x - this->x;
    float Y = y - this->y;
    float D = X+Y; // Manhatan distance
    float dx = X/D;
    float dy = Y/D;
    this->move(dx, dy, grid);
}

void unit::mine(struct square grid[GRID_WIDTH][GRID_HEIGHT]) {
    if(this->model->drill_power > 0 && !this->energy->is_empty()) {
        this->energy->decrease(1);
        for(int dx = -1; dx <= +1; dx++) {
            for(int dy = -1; dy <= +1; dy++) {
                int nx = (this->x + dx*SQUARE_SIZE/2) / SQUARE_SIZE;
                int ny = (this->y + dy*SQUARE_SIZE/2) / SQUARE_SIZE;
                if(nx >= 0 && nx < GRID_WIDTH && ny >= 0 && ny < GRID_HEIGHT) {
                    grid[nx][ny].mine(this);
                }
            }
        }
    }
}

void unit::build(struct model * model, float x, float y, std::forward_list<struct unit *> * units) {
    if(this->can_build(model)) {
        this->energy->decrease(model->energy_cost);
        this->cristal->decrease(model->cristal_cost);
        units->push_front(new unit(model, x, y));
    }
}

void unit::cycle(std::forward_list<struct unit *> * units, struct square grid[GRID_WIDTH][GRID_HEIGHT]) {
    if(this->model->generate_energy > 0 && !this->energy->is_full() && !this->powergem->is_empty()) {
        this->energy_accumulator += this->model->generate_energy * this->powergem->value;
        if(this->energy_accumulator >= ENERGY_GENERATOR_THRESHOLD) {
            this->energy->increase(1);
            this->energy_accumulator -= ENERGY_GENERATOR_THRESHOLD;
        }
    }

    if(this->model->collect_energy) {
        for(struct unit * unit : *units) {
            if(unit->model->give_energy && this->model != unit->model) {
                if(this->distance(unit) < COLLECT_DISTANCE) {
                    this->energy->gather(unit->energy);
                }
            }
        }
    }
    if(this->model->collect_cristal) {
        for(struct unit * unit : *units) {
            if(unit->model->give_cristal && this->model != unit->model) {
                if(this->distance(unit) < COLLECT_DISTANCE) {
                    this->cristal->gather(unit->cristal);
                }
            }
        }
    }
    if(this->model->collect_powergem) {
        for(struct unit * unit : *units) {
            if(unit->model->give_powergem && this->model != unit->model) {
                if(this->distance(unit) < COLLECT_DISTANCE) {
                    this->powergem->gather(unit->powergem);
                }
            }
        }
    }
    if(this->model->autobuild != NULL) {
        this->build(this->model->autobuild, this->x, this->y, units);
    }
    if(this->model->name == "midro") {
        if(target_x == 0.0 && target_y == 0.0) {
            target_x = 42;
            target_y = 42;
        }
    }
    if(this->target_x != 0.0 || this->target_y != 0.0) {
        this->move_toward(this->target_x, this->target_y, grid);
    }
}

float unit::distance(struct unit * unit) {
    return(this->distance(unit->x, unit->y));
}

float unit::distance(float x, float y) {
    return(abs(this->x - x) + abs(this->y - y));
}
